import java.util.*;
import java.io.*;

public class Align {
    
	// general information
    public static SpliceGraph s;
	
	public static Edge[] edge_list;
    public static String[] reads;
    public static ArrayList[] align;
    public static ArrayList[] confidence;
    public static double[] c;
	public static int fasta_readlen;
    
    public static void main(String[] args) {
		if (args.length != 3 && args.length != 4) {
			System.out.println("usage: Align <psg> <bowtie.align> <FASTA readlen> [<beta>]");
			System.exit(0);
		}
		
		s = new SpliceGraph(args[0]);
		
		String filename = args[1];
		fasta_readlen = Integer.parseInt(args[2]);
		process_reads(filename);
		
		try {
			s.setBeta(Double.parseDouble(args[3]));
		} catch (ArrayIndexOutOfBoundsException e) {
			// do nothing, this is okay
		}
		catch (NumberFormatException nfe) {
			System.out.print("Error: Beta should be given as a double.");
			System.exit(0);
		}
		
		// checking the edge case where we don't have any aligned reads
		if (reads.length == 0) {
			double ll = 0.0;
			//System.out.println("\tNumber of null reads: 0");
			System.out.println("\tNumber of reads: 0");
			s.emOutput(ll,-1);
			
			System.out.println();
			System.out.println("Current log likelihood: "+ll);
			System.out.println("Converged after 0 iterations");
			System.out.println("Log likelihood diff: 0");
			System.out.println("Max edge difference: 0.0");
			System.out.println("L1 norm: 0.0");
			System.out.println("L2 norm: 0.0");
			return;
		}
		
		// so if we're here, then we definitely have reads
		System.out.println("\tNumber of reads: "+reads.length);
		
		for (int i=0; i<align.length; i++) {
			for (int j=0; j<align[i].size(); j++) {
				Integer[] temp = (Integer[]) align[i].get(j);
				System.out.print("[");
				for (int k=0; k<temp.length; k++) {
					System.out.print(temp[k]);
					if (k<temp.length-1) {
						System.out.print(",");
					}
				}
				System.out.print("] ");
			}
			System.out.println();
		}
		
		double[] initial = s.getEdgeWeights();
		s.resetEdgeWeights();
		s.learnEM(reads,align,confidence,1);
		
		double[] a = s.getEdgeWeights();
		double result = 0;
		for (int i=0; i<initial.length; i++) {
			result += Math.abs(a[i]-initial[i]);
		}
		System.out.println("L1 norm: "+result);
		result = 0;
		for (int i=0; i<initial.length; i++) {
			result += Math.pow(a[i]-initial[i],2);
		}
		result = Math.sqrt(result);
		System.out.println("L2 norm: "+result);
    }
	
	// reads information from a bowtie alignment file
    private static void process_reads(String filename) {
		// initialize the temporary variables
		ArrayList<Integer> start = new ArrayList<Integer>();
		ArrayList<String> read = new ArrayList<String>();
		ArrayList<String> alig = new ArrayList<String>();
		ArrayList<String> which = new ArrayList<String>();
		ArrayList<String> mismatch = new ArrayList<String>();
		ArrayList<String> quality = new ArrayList<String>();
		
		// initialize the global variables
		reads = null;
		align = null;
		confidence = null;
		c = null;
		
		try {
			Scanner r = new Scanner(new File(filename));
			while (r.hasNextLine()) {
				String[] components = r.nextLine().split("\t");
				which.add(components[0]);
				alig.add(components[2]);
				start.add(Integer.decode(components[3]));
				read.add(components[4]);
				quality.add(components[5]);
				if (components.length == 8) {
					mismatch.add(components[7]);
				}
				else {
					mismatch.add("x");
				}
			}
		}
		catch (IOException ioe) {
			System.out.println("Alignment file "+filename+" contains an error.");
			System.exit(0);
		}
		
		c = compute_quality(quality,mismatch);
		reads = getReads(which,read);
		
		if (reads.length > 0) {
			process_align(reads.length,which,alig,c,reads[0].length(),start);
		}
    }
	
	// Computes confidence based on match/mismatch; quality of read
    private static double[] compute_quality(ArrayList<String> q, ArrayList<String> m) {
		double[] result = new double[q.size()];
		for (int i=0; i<q.size(); i++) {
			result[i] = 1; // initializing for multiplication
			// check for mismatches
			int[] positions = {-1};
			if (!m.get(i).equals("x")) {
				// process mismatches
				String[] mismatch = m.get(i).split(",");
				positions = new int[mismatch.length];
				for (int j=0; j<mismatch.length; j++) {
					// get position of mismatch - part before : should be position
					String[] temp = mismatch[j].split(":");
					positions[j] = Integer.parseInt(temp[0]);
				}
			}
			// process quality scores
			int index = 0; // for positions[], assume listed in increasing order
			for (int j=0; j<q.get(i).length(); j++) {
				int quality = (int) q.get(i).charAt(j);
				quality -= 33; // because we have a +33 offset
				if (index >= positions.length || j != positions[index]) {
					// not a mismatch
					result[i] *= 1-Math.pow(10,(-quality/10.0));
				}
				else {
					// mismatch
					result[i] *= (1/3.0)*Math.pow(10,(-quality/10.0));
					index++;
				}
			}
		}
		return result;
    }
	
	// retrieves the reads from the read arraylist
	private static String[] getReads(ArrayList<String> r, ArrayList<String> s) {
		int num = count_unique(r);
		String[] result = new String[num];
		String id = null;
		int index = 0;
		ArrayList<String> used_ids = new ArrayList<String>();
		for (int i=0; i<r.size(); i++) {
			if (!contains(used_ids, r.get(i))) {
				id = r.get(i);
				used_ids.add(id);
				result[index] = s.get(i);
				index++;
			}
		}
		return trim(result);
    }
	
	private static String[] trim(String[] to_trim) {
		int i=0;
		for (i=0; i<to_trim.length; i++) {
			if (to_trim[i] == null) {
				break;
			}
		}
		String[] result = new String[i];
		for (i=0; i<result.length; i++) {
			result[i] = to_trim[i];
		}
		return result;
    }
	
	// Assumption: any duplicates will be consecutive
    private static int count_unique(ArrayList<String> r) {
		if (r.size() == 0) { return 0; }
		int unique = 1;
		int index = 1;
		String current = r.get(0);
		while (index < r.size()) {
			if (!r.get(index).equals(current)) {
				current = r.get(index);
				unique++;
			}
			index++;
		}
		return unique;
    }
	
	private static boolean contains(ArrayList<String> a, String b) {
		boolean retval = false;
		for (int i=0; i<a.size(); i++) {
			if (a.get(i).equals(b)) {
				retval = true;
			}
		}
		return retval;
    }
    
    private static boolean contains(ArrayList<Integer[]> a, Integer[] b) {
		boolean retval = false;
		for (int i=0; i<a.size(); i++) {
			if (a.get(i).length == b.length) {
				boolean verify = true;
				for (int j=0; j<a.get(i).length; j++) {
					if (a.get(i)[j] != b[j]) {
						verify = false;
					}
				}
				if (verify) {
					retval = true;
					break;
				}
			}
		}
		return retval;
    }
	
	private static void process_align(int l, ArrayList<String> id, 
									  ArrayList<String> al, double[] c,
									  int read_len, ArrayList<Integer> st) {
		align = new ArrayList[l];
		confidence = new ArrayList[l];
		
		int index = 0;
		int old_ind = -1;
		ArrayList<Integer[]> matches = new ArrayList<Integer[]>();
		ArrayList<Double> quality = new ArrayList<Double>();
		
		HashMap<String,Integer> used_ids = new HashMap<String,Integer>();
		
		for (int i=0; i<id.size(); i++) {
			if (!used_ids.containsKey(id.get(i))) {
				//used_ids.put(id.get(i), index);
				matches = new ArrayList<Integer[]>();
				quality = new ArrayList<Double>();
			}
			else {
				old_ind = used_ids.get(id.get(i)).intValue();
				matches = (ArrayList<Integer[]>) align[old_ind];
				quality = (ArrayList<Double>) confidence[old_ind];
			}
			String[] match = al.get(i).substring(1).split(":");
			Integer[] r = new Integer[match.length];
			for (int j=0; j<match.length; j++) {
				r[j] = Integer.decode(match[j]);
			}
			// check whether this alignment needs to be trimmed
			// and do so if necessary
			if (r.length == 1) {
				// st.get(i) is an index into that segment
				// do nothing, this is as simple as it gets
			}
			else {
				// find first segment used
				int j=0;
				int pos = st.get(i).intValue();
				Segment temp_1 = s.getSegment(r[0]);
				if (pos > Math.min(temp_1.length(),fasta_readlen)) {
					pos -= Math.min(temp_1.length(),fasta_readlen-1);
					j++;
					temp_1 = s.getSegment(r[j]);
					while (pos > temp_1.length()) {
						if (temp_1.length() > 0) {
							pos -= Math.min(temp_1.length(),fasta_readlen-1);
							j++;
							temp_1 = s.getSegment(r[j]);
						}
						else {
							pos = 0;
						}
					}
				}
				
				// pos is now an index into the jth segment in r
				// (subsegment if j=0)
				int total = 0;
				if (j == 0) {
					total += Math.min(s.getSegment(r[0]).length(),fasta_readlen-1)-pos;
				}
				else {
					total += s.getSegment(r[j]).length()-pos;
				}
				int k=j;
				if (total >= read_len || s.getSegment(r[j]).length() == 0) {
					// then j is the only relevant segment
				}
				else {
					// let's find the end!
					while (total < read_len) {
						k++;
						temp_1 = s.getSegment(r[k]);
						if (temp_1.length() > 0) {
							total += temp_1.length();
						}
						else {
							// we've arrived at the poly-A tail!
							total = read_len;
						}
					}
				}
				
				// trim r between j and k
				Integer[] temp = new Integer[k-j+1];
				int ind = 0;
				for (int x=j; x<=k; x++) {
					temp[ind] = r[x];
					ind++;
				}
				r = temp;
				
			}
			if (!contains(matches,r)) {
				matches.add(r);
				quality.add(c[i]);
			}
			if (!used_ids.containsKey(id.get(i))) {
				used_ids.put(id.get(i), index);
				align[index] = matches;
				confidence[index] = quality;
				index++;
			}
			else {
				align[old_ind] = matches;
				confidence[old_ind] = quality;
			}
		}
    }	
}