import re
import sys
import subprocess
import os
import itertools
import splicegraph
import gzip
import string
import table
import gff
import fasta
import fastq
import collections
import shutil

class Reference:
    def __init__(self, ref_name):
        index_filename = os.path.join(ref_name, "index.txt")
        self.genes = list(table.FileReader(index_filename))
        self.genes.sort(key=lambda gene: gene["gene_id"])
        self.gene_id_to_chrom_map = dict([(gene["gene_id"], gene["chrom"])
                                          for gene in self.genes])

def learnpsg_classpath():
    return os.path.join(os.path.abspath(os.path.dirname(__file__)), "LearnPSG")

class Error(Exception):
    pass

# Returns the length of the first read in the given FASTQ file, or None if the file contains zero reads
def fastq_read_length(fastq_filename):
    with open(fastq_filename) as fastq_file:
        for read in fastq.Reader(fastq_file):
            return len(read.sequence.rstrip())
    return None

def reference_read_length_filename(ref_name):
    return os.path.join(ref_name, "max_read_length")

def write_reference_read_length(ref_name, max_read_length):
    with open(reference_read_length_filename(ref_name), 'w') as read_length_file:
        print >>read_length_file, max_read_length

def reference_read_length(ref_name):
    with open(reference_read_length_filename(ref_name)) as read_length_file:
        return int(read_length_file.read())

def alignments_by_gene(align_filename):
    id_field_num = 2
    stream = gzip.GzipFile(align_filename)
    curr_gene_id = None
    curr_alignments = []
    for line in stream:
        tokens = line.split('\t')
        global_sequence_id = tokens[id_field_num]
        gene_id, sequence_id = global_sequence_id.rsplit('.', 1)

        tokens[id_field_num] = sequence_id
        
        if gene_id != curr_gene_id:
            if curr_alignments:
                yield curr_gene_id, curr_alignments
            curr_gene_id = gene_id
            curr_alignments = []

        curr_alignments.append(tokens)

    if curr_alignments:
        yield curr_gene_id, curr_alignments

    stream.close()

def gene_alignments(genes, align_filename):
    alignments_iterator = alignments_by_gene(align_filename)
    next_gene, next_alignments = next(alignments_iterator, (None, None))
    for gene in genes:
        gene_id = gene["gene_id"]
        if gene_id != next_gene:
            yield []
        else:
            yield next_alignments
            next_gene, next_alignments = next(alignments_iterator, (None, None))
    if next_gene is not None:
        raise Error("Alignments were not sorted by gene")

def flatten(list_of_lists):
    result = []
    for lst in list_of_lists:
        result.extend(lst)
    return result
    
def gene_joined_alignments(genes, align_filenames):
    all_alignments = itertools.izip(*[gene_alignments(genes, align_filename)
                                      for align_filename in align_filenames])
    for alignments in all_alignments:
        yield flatten(alignments)
    
def split_alignment(genes, align_filename, end, out_dir, gene_id_to_chrom_map,
                    append=False):
    for gene, alignments in itertools.izip(genes, gene_alignments(genes,
                                                                  align_filename)):
        gene_id = gene["gene_id"]
        chrom = gene_id_to_chrom_map[gene_id]
        chrom_dir = os.path.join(out_dir, chrom)
        gene_filename = os.path.join(chrom_dir,
                                     "%s.%d.align" % (gene_id, end))
        gene_file = open(gene_filename, 'a' if append else 'w')
        for alignment in sorted(alignments):
            gene_file.write('\t'.join(alignment))
        gene_file.close()

def parsed_em_stream(stream):
    def parse_field(line, pat, value_type):
        m = pat.match(line)
        if m is None:
            raise Error("Could not match line with pat:\n%s\n%s" % (line, pat))
        return value_type(m.group(1))
    
    lines = list(stream)
    d = {}
    fp = r'([-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)'
    beta_pat = re.compile(r'beta = ' + fp)
    num_reads_pat = re.compile(r'\tNumber of reads: (\d+)')
    log_likelihood_pat = re.compile(r'Current log likelihood: ' + fp)
    iterations_pat = re.compile(r'Converged after (\d+) iterations')
    log_likelihood_diff_pat = re.compile(r'Log likelihood diff: ' + fp)
    max_edge_diff_pat = re.compile(r'Max edge difference: ' + fp)
    l1_norm_pat = re.compile(r'L1 norm: ' + fp)
    l2_norm_pat = re.compile(r'L2 norm: ' + fp)

    if lines[0].startswith("beta"): 
        d["beta"] = parse_field(lines[0], beta_pat, float)
        offset = 1
    else:
        offset = 0

    d["num_reads"] = parse_field(lines[offset], num_reads_pat, int)
    d["log_likelihood"] = parse_field(lines[-6], log_likelihood_pat, float)
    d["num_iterations"] = parse_field(lines[-5], iterations_pat, int)
    d["log_likelihood_diff"] = parse_field(lines[-4], log_likelihood_diff_pat, float)
    d["max_edge_diff"] = parse_field(lines[-3], max_edge_diff_pat, float)
    d["l1_norm"] = parse_field(lines[-2], l1_norm_pat, float)
    d["l2_norm"] = parse_field(lines[-1], l2_norm_pat, float)

    iterations = []
    for line in lines[offset + 1:-7]:
        tokens = line.split()
        log_likelihood = float(tokens[1])
        parameters = map(float, tokens[3:])
        iterations.append((log_likelihood, parameters))
    d["iterations"] = iterations
    return d
    
def parsed_em_output(filename):
    with open(filename) as f:
        return parsed_em_stream(f)

def infer_frequencies(genes,
                      alignments1,
                      alignments2,
                      sample_name,
                      ref_name,
                      fragment_length_mean,
                      fragment_length_stddev,
                      beta,
                      pool=None,
                      tempdir=None):
    read_length = reference_read_length(ref_name)

    if tempdir is None:
        tempdir = sample_name
    
    gene_results_filename = os.path.join(sample_name, "gene.results.txt")
    gene_results_fields = ["gene_id",
                           "chrom",
                           "start",
                           "end",
                           "strand",
                           "num_transcripts",
                           "num_paths",
                           "num_parameters",
                           "num_reads",
                           "log_likelihood",
                           "parameters"]

    gene_results_debug_filename = os.path.join(sample_name, "gene.results.debug.txt")
    gene_results_debug_fields = (gene_results_fields +
                                 ["num_iterations",
                                  "log_likelihood_diff",
                                  "max_edge_diff",
                                  "l1_norm",
                                  "l2_norm"])
    gene_results_debug_writer = table.FileWriter(gene_results_debug_filename,
                                                 gene_results_debug_fields)

    inferrer = PSGFrequencyInferrer(sample_name,
                                    ref_name,
                                    read_length,
                                    fragment_length_mean,
                                    fragment_length_stddev,
                                    beta,
                                    tempdir)

    gene_alignments = itertools.izip(genes, alignments1, alignments2)
    
    if pool is not None:
        iterator = pool.imap_unordered(inferrer, gene_alignments, chunksize=10)
    else:
        iterator = itertools.imap(inferrer, gene_alignments)

    results = []
    for result in iterator:
        log(result["command"])
        if not result["success"]:
            log("Error: AlignPaired failed on gene %s with message: %s" %
                (result["gene_id"], result["error_message"]))
            continue
        if result["log"]:
            log("Warning: AlignPaired generated log info for gene %s: %s" %
                (result["gene_id"], result["log"]))
        gene_results_debug_writer.write(result)
        results.append(result)
    gene_results_debug_writer.close()

    # Sort results by gene_id
    results.sort(key=lambda r: r["gene_id"])

    # Output final gene results table
    gene_results_writer = table.FileWriter(gene_results_filename,
                                           gene_results_fields)
    for result in results:
        gene_results_writer.write(result)
    gene_results_writer.close()

    # Output isoform results table
    isoform_results_filename = os.path.join(sample_name, "isoform.results.txt")
    isoform_results_fields = ["gene_id",
                              "transcript_id",
                              "prob"]
    isoform_results_writer = table.FileWriter(isoform_results_filename,
                                              isoform_results_fields)
    for result in results:
        for transcript_id, prob in result["isoform_probs"]:
            isoform_results_writer.write({"gene_id": result["gene_id"],
                                          "transcript_id": transcript_id,
                                          "prob": prob})
    isoform_results_writer.close()

def create_empty_file(filename):
    f = open(filename, 'w')
    f.close()

def log(s):
    print >>sys.stderr, s
        
def run_command(cmd):
    try:
        subprocess.check_call(cmd, shell=True)
    except subprocess.CalledProcessError, e:
        raise Error("Command was terminated or failed:\n%s" % cmd)

def write_alignments(filename, alignments):
    with open(filename, 'w') as alignment_file:
        for alignment in sorted(alignments):
            alignment_file.write('\t'.join(alignment))
    
def infer_gene_frequencies(gene,
                           alignments1,
                           alignments2,
                           sample_name,
                           ref_name,
                           read_length,
                           fragment_length_mean,
                           fragment_length_stddev,
                           beta,
                           tempdir):
    gene_id = gene["gene_id"]
    chrom = gene["chrom"]
    psg_src_filename = os.path.join(ref_name, chrom, gene_id + ".psg")
    psg_filename = os.path.join(tempdir, gene_id + ".psg")
    shutil.copy(psg_src_filename, psg_filename)
    align1_filename = os.path.join(tempdir, gene_id + ".1.align")
    align2_filename = os.path.join(tempdir, gene_id + ".2.align")
    write_alignments(align1_filename, alignments1)
    write_alignments(align2_filename, alignments2)
            
    em_filename = os.path.join(tempdir, gene_id + ".em.out")
    em_log_filename = em_filename + ".log"
    cmd = ("java -cp %s AlignPaired %s %s %s %d %d %f %f > %s 2> %s" %
           (learnpsg_classpath(),
            psg_filename,
            align1_filename,
            align2_filename,
            read_length,
            fragment_length_mean,
            fragment_length_stddev,
            beta,
            em_filename,
            em_log_filename))

    result = gene.copy()
    result["command"] = cmd
    try:
        run_command(cmd)
        with open(em_log_filename) as log_file:
            result["log"] = log_file.read()
        result.update(parsed_em_output(em_filename))
        result["success"] = True

        edge_probs = result["iterations"][-1][1]
        result["parameters"] = ",".join(["%.3g" % p for p in edge_probs])

        # Create PSG with learned parameters
        psg_file = open(psg_filename)
        psg = splicegraph.read_splice_graph(psg_file)
        psg_file.close()
        assert(len(psg.edges) == len(edge_probs))
        psg.setEdgeProbs(edge_probs)
        psg.normalize()

        # Write PSG file with learned parameters
        # sample_psg_filename = os.path.join(sample_name, chrom, gene_id + ".psg")
        # sample_psg_file = open(sample_psg_filename, 'w')
        # splicegraph.write_splice_graph(sample_psg_file, psg)
        # sample_psg_file.close()

        # Compute path probabilities of annotated transcripts
        isoform_probs = []
        psg_paths_filename = os.path.join(ref_name, chrom, gene_id + ".paths")
        psg_paths_file = open(psg_paths_filename)
        for line in psg_paths_file:
            transcript_id, path_string = line.split('\t')
            path = map(int, path_string.split())
            path_prob = psg.subpathProb(path)
            isoform_probs.append((transcript_id, path_prob))
        psg_paths_file.close()
        result["isoform_probs"] = isoform_probs

        # Remove input and output files
        os.remove(psg_filename)
        os.remove(align1_filename)
        os.remove(align2_filename)
        os.remove(em_filename)
        os.remove(em_log_filename)
    except Error, e:
        result["success"] = False
        result["error_message"] = e.args[0]
    
    return result

class PSGFrequencyInferrer:
    def __init__(self, sample_name, ref_name,
                 read_length, fragment_length_mean, fragment_length_stddev, beta,
                 tempdir):
        self.sample_name = sample_name
        self.ref_name = ref_name
        self.read_length = read_length
        self.fragment_length_mean = fragment_length_mean
        self.fragment_length_stddev = fragment_length_stddev
        self.beta = beta
        self.tempdir = tempdir
    def __call__(self, (gene, alignments1, alignments2)):
        return infer_gene_frequencies(gene,
                                      alignments1,
                                      alignments2,
                                      self.sample_name,
                                      self.ref_name,
                                      self.read_length,
                                      self.fragment_length_mean,
                                      self.fragment_length_stddev,
                                      self.beta,
                                      self.tempdir)

class Transcript:
    def __init__(self, transcript_id, gene_id, chrom, strand, exons=None):
        self.transcript_id = transcript_id
        self.gene_id = gene_id
        self.chrom = chrom
        self.strand = strand
        self.exons = exons if exons is not None else []

reverse_complement_translate_table = string.maketrans("ACGTNacgtn",
                                                      "TGCANtgcan")
def reverse_complement(s):
    return ''.join(reversed(s.translate(reverse_complement_translate_table)))

def chrom_sequence_from_dir(chrom, genome_dir):
    fasta_extensions = ["fa", "fasta"]
    for extension in fasta_extensions:
        filename = os.path.join(genome_dir, chrom + "." + extension)
        if os.path.exists(filename):
            fasta_file = open(filename)
            recs = list(fasta.Reader(fasta_file))
            if len(recs) != 1:
                raise Error("Chromosome FASTA file does not " +
                            "contain exactly one sequence: " +
                            filename)
            return recs[0].sequence
    else:
        raise Error((("Could not locate a FASTA file for chromosome '%s' " +
                      "in directory '%s'") % (chrom, genome_dir)))
        
class FastaDirectoryGenomeDB:
    def __init__(self, directory, cache_last_chrom=False):
        self.directory = directory
        self.cache_last_chrom = cache_last_chrom
        self.last_chrom = None
        self.last_chrom_sequence = None

    def sequence(self, chrom, start, end, strand):
        if self.last_chrom != chrom:
            chrom_sequence = chrom_sequence_from_dir(chrom, self.directory)
            if self.cache_last_chrom:
                self.last_chrom = chrom
                self.last_chrom_sequence = chrom_sequence
        else:
            chrom_sequence = self.last_chrom_sequence

        seq = chrom_sequence[start: end]
        return seq if strand == '+' else reverse_complement(seq)
        
def transcripts_from_gtf(filename):
    transcripts = {}
    gtf_file = open(filename)
    for rec in gff.Reader(gtf_file):
        if rec.type == "exon":
            try:
                transcript_id = rec.attributes["transcript_id"][0]
                if transcript_id not in transcripts:
                    gene_id = rec.attributes["gene_id"][0]
                    transcripts[transcript_id] = Transcript(transcript_id,
                                                            gene_id,
                                                            rec.seqid,
                                                            rec.strand)
                exon = (rec.start - 1, rec.end)
                transcripts[transcript_id].exons.append(exon)
            except KeyError, e:
                raise Error(("exon record is missing required " +
                             "attribute '%s'\n%s") % (e.args[0], repr(rec)))
    return transcripts.values()

def transcripts_grouped_by_gene(transcripts):
    by_gene = collections.defaultdict(list)
    for transcript in transcripts:
        by_gene[transcript.gene_id].append(transcript)
    return by_gene

def genes_grouped_by_chrom(genes):
    by_chrom = collections.defaultdict(dict)
    for gene_id, gene_transcripts in genes.items():
        by_chrom[gene_transcripts[0].chrom][gene_id] = gene_transcripts
    return by_chrom
