#!/usr/bin/env python

import sys
import optparse
import collections
import os.path
import os
import subprocess
import multiprocessing
import itertools
import signal

import gff
import splicegraph
import fasta
import string
import psg
import table
import chisquare
import fdr

def init_worker():
    signal.signal(signal.SIGINT, signal.SIG_IGN)

def sample_results(sample):
    filename = os.path.join(sample, "gene.results.txt")
    return dict([(r["gene_id"], r) for r in table.FileReader(filename)])
    
def chisquare_p(d, df):
    return chisquare.lchisqprob(d, df)

def sample_gene_mle_params(result):
    return map(float, result["parameters"].split(","))

def compute_max_delta_psi(result1, result2):
    mle_params = map(sample_gene_mle_params, [result1, result2])
    return max([abs(p1 - p2) for p1, p2 in zip(*mle_params)])

def test_diff_splicing(genes, sample1_name, sample2_name, comp_name):
    r1 = sample_results(sample1_name)
    r2 = sample_results(sample2_name)
    rc = sample_results(comp_name)
    gene_ids_in_all = set(r1).intersection(set(r2)).intersection(set(rc))

    results = []
    for gene in genes:
        gene_id = gene["gene_id"]
        if gene_id not in gene_ids_in_all:
            print >>sys.stderr, "Warning: Gene ID not in all samples:", gene_id
            continue
        result = dict(gene)
        result["max_delta_psi"] = compute_max_delta_psi(r1[gene_id], r2[gene_id])
        result["sample1_num_reads"] = int(r1[gene_id]["num_reads"])
        result["sample2_num_reads"] = int(r2[gene_id]["num_reads"])
        result["joint_num_reads"] = int(rc[gene_id]["num_reads"])
        result["sample1_ll"] = float(r1[gene_id]["log_likelihood"])
        result["sample2_ll"] = float(r2[gene_id]["log_likelihood"])
        result["diff_ll"] = result["sample1_ll"] + result["sample2_ll"]
        result["null_ll"] = float(rc[gene_id]["log_likelihood"])
        result["lrt_d"] = 2 * (result["diff_ll"] - result["null_ll"])
        result["df"] = int(gene["num_parameters"])
        if result["df"] > 0:
            result["p_value"] = chisquare_p(result["lrt_d"], result["df"])
        else:
            result["p_value"] = "NA"
            result["fdr"] = "NA"
        results.append(result)

    # Compute FDR values
    results_with_p_values = [r for r in results if r["p_value"] != "NA"]
    p_values = [r["p_value"] for r in results_with_p_values]
    fdrs = fdr.benjamini_hochberg(p_values)
    for result, result_fdr in zip(results_with_p_values, fdrs):
        result["fdr"] = result_fdr

    # Output results
    diff_filename = os.path.join(comp_name, "diff.txt")
    field_names = ["gene_id", "chrom", "start", "end", "strand",
                   "num_transcripts", "num_paths", "num_parameters",
                   "sample1_num_reads", "sample2_num_reads", "joint_num_reads",
                   "max_delta_psi",
                   "sample1_ll", "sample2_ll", "diff_ll",
                   "null_ll", "lrt_d", "df", "p_value", "fdr"]
    diff_writer = table.FileWriter(diff_filename, field_names)
    for result in results:
        diff_writer.write(result)
    diff_writer.close()
        
def main():
    usage = "usage: %prog [options] comp_name sample1_name sample2_name reference_name"
    parser = optparse.OptionParser(usage=usage)
    parser.add_option("--frag-mean", dest="fragment_length_mean", 
                      help="Fragment length mean (default: 200)",
                      type="int", default=200, metavar="MEAN")
    parser.add_option("--frag-stddev", dest="fragment_length_stddev",
                      help="Fragment length standard deviation (default: 25)",
                      type="float", default=25.0, metavar="STDDEV")
    parser.add_option("--pseudocount", dest="beta",
                      help="Prior pseudocount for each edge.  Set this to"
                      "zero for maximum likelihood estimates.  Set to a "
                      "positive value to obtain smoother estimates when the "
                      "number of reads is small (default: 1)",
                      type="float", default=1, metavar="COUNT")
    parser.add_option("--tempdir", dest="tempdir",
                      help="Temporary directory used for intermediate files"
                      "(default: comp_name)",
                      type="str", default=None, metavar="DIR")
    options, args = parser.parse_args()
    if len(args) != 4:
        parser.error("Invalid number of arguments")
    comp_name, sample1_name, sample2_name, ref_name = args

    num_cores = multiprocessing.cpu_count()
    print >>sys.stderr, "Using %d cores" % num_cores
    pool = multiprocessing.Pool(num_cores, init_worker)

    try:
        if not os.path.exists(comp_name):
            os.makedirs(comp_name)

        reference = psg.Reference(ref_name)
        
        alignments = []
        for end in (1, 2):
            filenames = [os.path.join(sample, "alignment%d.bowtie.gz" % end)
                         for sample in sample1_name, sample2_name]
            alignments.append(psg.gene_joined_alignments(reference.genes,
                                                         filenames))
                
        # Run PSG inferrer
        psg.infer_frequencies(reference.genes,
                              alignments[0],
                              alignments[1],
                              comp_name,
                              ref_name,
                              options.fragment_length_mean,
                              options.fragment_length_stddev,
                              options.beta,
                              pool,
                              options.tempdir)
        
        # Perform differential splicing test
        psg.log("Performing differential splicing test")
        test_diff_splicing(reference.genes, sample1_name, sample2_name, comp_name)
        
    except psg.Error, e:
        print >>sys.stderr, "Error: %s" % e.args[0]
        sys.exit(1)
    except KeyboardInterrupt:
        print >>sys.stderr, "Caught KeyboardInterrupt, terminating workers"
        pool.terminate()
        pool.join()
        sys.exit(1)

if __name__ == "__main__":
    main()
