#!/usr/bin/env python

import sys
import optparse
import collections
import os.path
import os
import subprocess
import multiprocessing
import itertools
import signal
import time

import gff
import splicegraph
import fasta
import string
import psg
import table

class WallClockTimer:
    def __init__(self):
        self.last_time = time.time()
    def elapsed_time(self):
        return time.time() - self.last_time
    def reset(self):
        old_last_time = self.last_time
        self.last_time = time.time()
        return self.last_time - old_last_time

def init_worker():
    signal.signal(signal.SIGINT, signal.SIG_IGN)

def run_bowtie(reads_filename, ref_name, align_filename, num_cores,
               temp_dir=None, sort_buffer_size=None, trim=None):
    bowtie_prefix = os.path.join(ref_name, "references")
    cmd = ("bowtie --best --strata -p %d -l 25 -a -m 200 %s %s 2> %s" %
           (num_cores,
            bowtie_prefix,
            reads_filename,
            align_filename + ".log"))
    sort_cmd = "sort -k3,3"
    if temp_dir is not None:
        sort_cmd += " -T " + temp_dir
    if sort_buffer_size is not None:
        sort_cmd += " -S " + sort_buffer_size
    cmd += " | " + sort_cmd
    if trim:
        cmd += " | ./trim_bowtie_alignments.py %d" % trim
    cmd += " | gzip -c > " + align_filename
    psg.log(cmd)
    psg.run_command(cmd)

def main():
    usage = "usage: %prog [options] reference_name sample_name upstream_read_file downstream_read_file"
    parser = optparse.OptionParser(usage=usage)
    parser.add_option("--frag-mean", dest="fragment_length_mean", 
                      help="Fragment length mean (default: 200)",
                      type="int", default=200, metavar="MEAN")
    parser.add_option("--frag-stddev", dest="fragment_length_stddev",
                      help="Fragment length standard deviation (default: 25)",
                      type="float", default=25.0, metavar="STDDEV")
    parser.add_option("--pseudocount", dest="beta",
                      help="Prior pseudocount for each edge.  Set this to "
                      "zero for maximum likelihood estimates.  Set to a "
                      "positive value to obtain smoother estimates when the "
                      "number of reads is small (default: 1)",
                      type="float", default=1, metavar="COUNT")
    parser.add_option("--tempdir", dest="tempdir",
                      help="Temporary directory used for sorting alignments "
                      "(default: use default for 'sort' command)",
                      type="str", default=None, metavar="DIR")
    parser.add_option("--sort-buffer-size", dest="sort_buffer_size",
                      help="Buffer size in memory for sorting alignments "
                      "(default: use default for 'sort' command)",
                      default=None, metavar="SIZE")
    parser.add_option("--trim-length", dest="trim",
                      help="Length to trim reads to after alignment"
                      "(default: no trimming)",
                      type="int", default=None, metavar="LENGTH")
    options, args = parser.parse_args()
    if len(args) != 4:
        parser.error("Invalid number of arguments")
    sample_name, ref_name, end1_filename, end2_filename = args

    num_cores = multiprocessing.cpu_count()
    print >>sys.stderr, "Using %d cores" % num_cores
    pool = multiprocessing.Pool(num_cores, init_worker)

    try:
        timer = WallClockTimer()
        
        if not os.path.exists(sample_name):
            os.makedirs(sample_name)

        # Read reference index
        index_filename = os.path.join(ref_name, "index.txt")
        genes = list(table.FileReader(index_filename))
        genes.sort(key=lambda gene: gene["gene_id"])
        gene_id_to_chrom_map = {}
        for rec in genes:
            gene_id_to_chrom_map[rec["gene_id"]] = rec["chrom"]
        reference_read_time = timer.reset()
        psg.log("Reference reading time: %.0f seconds" % reference_read_time)

        # Run Bowtie
        end1_align_filename = os.path.join(sample_name, "alignment1.bowtie.gz")
        end2_align_filename = os.path.join(sample_name, "alignment2.bowtie.gz")
        run_bowtie(end1_filename, ref_name, end1_align_filename, num_cores,
                   options.tempdir, options.sort_buffer_size, options.trim)
        alignment1_time = timer.reset()
        run_bowtie(end2_filename, ref_name, end2_align_filename, num_cores,
                   options.tempdir, options.sort_buffer_size, options.trim)
        alignment2_time = timer.reset()
        alignment_time = alignment1_time + alignment2_time
        psg.log("Mate 1 alignment time: %.0f seconds" % alignment1_time)
        psg.log("Mate 2 alignment time: %.0f seconds" % alignment2_time)
        psg.log("Total alignment time: %.0f seconds" % alignment_time)

        # Create alignment files for each PSG
        alignments = [psg.gene_alignments(genes, align_filename)
                      for align_filename in
                      [end1_align_filename, end2_align_filename]]

        # Run PSG inferrer
        psg.infer_frequencies(genes,
                              alignments[0],
                              alignments[1],
                              sample_name,
                              ref_name,
                              options.fragment_length_mean,
                              options.fragment_length_stddev,
                              options.beta,
                              pool,
                              options.tempdir)
        estimation_time = timer.reset()
        total_time = alignment_time + estimation_time + reference_read_time
        psg.log("Estimation time: %.0f seconds" % estimation_time)
        psg.log("Total time: %.0f seconds" % total_time)
        
    except psg.Error, e:
        print >>sys.stderr, "Error: %s" % e.args[0]
        sys.exit(1)
    except KeyboardInterrupt:
        print >>sys.stderr, "Caught KeyboardInterrupt, terminating workers"
        pool.terminate()
        pool.join()
        sys.exit(1)

if __name__ == "__main__":
    main()
