#!/usr/bin/env python

import sys
import optparse
import collections
import os.path
import os
import subprocess
import multiprocessing
import itertools

import gff
import splicegraph
import fasta
import string
import table
import psg

class Transcript:
    def __init__(self, transcript_id, gene_id, chrom, strand, exons=None):
        self.transcript_id = transcript_id
        self.gene_id = gene_id
        self.chrom = chrom
        self.strand = strand
        self.exons = exons if exons is not None else []

reverse_complement_translate_table = string.maketrans("ACGTNacgtn",
                                                      "TGCANtgcan")
def reverse_complement(s):
    return ''.join(reversed(s.translate(reverse_complement_translate_table)))

def chrom_sequence_from_dir(chrom, genome_dir):
    fasta_extensions = ["fa", "fasta"]
    for extension in fasta_extensions:
        filename = os.path.join(genome_dir, chrom + "." + extension)
        if os.path.exists(filename):
            fasta_file = open(filename)
            recs = list(fasta.Reader(fasta_file))
            if len(recs) != 1:
                raise psg.Error("Chromosome FASTA file does not " +
                                "contain exactly one sequence: " +
                                filename)
            return recs[0].sequence
    else:
        raise psg.Error((("Could not locate a FASTA file for chromosome '%s' " +
                          "in directory '%s'") % (chrom, genome_dir)))
        
class FastaDirectoryGenomeDB:
    def __init__(self, directory, cache_last_chrom=False):
        self.directory = directory
        self.cache_last_chrom = cache_last_chrom
        self.last_chrom = None
        self.last_chrom_sequence = None

    def sequence(self, chrom, start, end, strand):
        if self.last_chrom != chrom:
            chrom_sequence = chrom_sequence_from_dir(chrom, self.directory)
            if self.cache_last_chrom:
                self.last_chrom = chrom
                self.last_chrom_sequence = chrom_sequence
        else:
            chrom_sequence = self.last_chrom_sequence

        seq = chrom_sequence[start: end]
        return seq if strand == '+' else reverse_complement(seq)

def read_transcripts(filename):
    if filename is None:
        psg.log("Reading transcripts from GTF-formatted standard input...")
        return transcripts_from_gtf_stream(sys.stdin)
    else:
        psg.log("Reading transcripts from GTF-formatted file...")
        with open(filename) as stream:
            return transcripts_from_gtf_stream(stream)
    
def transcripts_from_gtf_stream(stream):
    transcripts = {}
    for rec in gff.Reader(stream):
        if rec.type == "exon":
            try:
                transcript_id = rec.attributes["transcript_id"][0]
                if transcript_id not in transcripts:
                    gene_id = rec.attributes["gene_id"][0]
                    transcripts[transcript_id] = Transcript(transcript_id,
                                                            gene_id,
                                                            rec.seqid,
                                                            rec.strand)
                exon = (rec.start - 1, rec.end)
                transcripts[transcript_id].exons.append(exon)
            except KeyError, e:
                raise MyError(("exon record is missing required " +
                             "attribute '%s'\n%s") % (e.args[0], repr(rec)))
    return transcripts.values()

def transcripts_grouped_by_gene(transcripts):
    by_gene = collections.defaultdict(list)
    for transcript in transcripts:
        by_gene[transcript.gene_id].append(transcript)
    return by_gene

def genes_grouped_by_chrom(genes):
    by_chrom = collections.defaultdict(dict)
    for gene_id, gene_transcripts in genes.items():
        by_chrom[gene_transcripts[0].chrom][gene_id] = gene_transcripts
    return by_chrom

class PSGCreator:
    def __init__(self, chrom, db, read_length, k, unfactorized, output_dir):
        self.chrom = chrom
        self.db = db
        self.read_length = read_length
        self.output_dir = output_dir
        self.k = k
        self.unfactorized = unfactorized
    def __call__(self, gene):
        gene_id = gene[0]
        transcripts = sorted(gene[1], key=lambda t: t.transcript_id)
        
        log = ""
        
        exon_lists = [sorted(transcript.exons) for transcript in transcripts]
        start = min([exon_list[0][0] for exon_list in exon_lists])
        end = max([exon_list[-1][1] for exon_list in exon_lists])
        strand = transcripts[0].strand

        # Create initial splice graph
        (gene_splice_graph, intervals, transcript_paths) = \
            splicegraph.kth_order_splice_graph(self.chrom,
                                               strand,
                                               exon_lists,
                                               self.db,
                                               self.k)
        if self.unfactorized:
            if gene_splice_graph.getNumIsoforms() > 1000:
                log += "Warning: Did not unfactorize gene because it has more than 1000 paths: " + gene_id + "\n"
            else:
                gene_splice_graph, transcript_paths = splicegraph.unfactored_splice_graph_and_paths(gene_splice_graph,
                                                                                                    transcript_paths)
        
        splice_graph_filename = os.path.join(self.output_dir, gene_id + ".psg")
        splice_graph_file = open(splice_graph_filename, 'w')
        splicegraph.write_splice_graph(splice_graph_file, gene_splice_graph)
        splice_graph_file.close()

        transcript_paths_filename = os.path.join(self.output_dir, gene_id + ".paths")
        transcript_paths_file = open(transcript_paths_filename, 'w')
        for transcript, path in zip(transcripts, transcript_paths):
            line = "%s\t%s" % (transcript.transcript_id,
                               ' '.join(map(str, path)))
            print >>transcript_paths_file, line
        transcript_paths_file.close()
        
        # Make FASTA file for this splice graph
        cmd = ("java -cp %s MakeFastaStdOut %s %d" %
               (psg.learnpsg_classpath(),
                splice_graph_filename,
                self.read_length))
        log += cmd + "\n"
        make_fasta_process = subprocess.Popen(cmd, shell=True,
                                              stdout=subprocess.PIPE)
        make_fasta_out = make_fasta_process.stdout
        raw_fasta_records = list(fasta.Reader(make_fasta_out))
        make_fasta_out.close()
        make_fasta_process.wait()
        if make_fasta_process.returncode != 0:
            raise psg.Error("Command was terminated or failed:\n%s" % cmd)

        raw_fasta_filename = os.path.join(self.output_dir, gene_id + ".fasta")
        raw_fasta_file = open(raw_fasta_filename, 'w')
        raw_fasta_writer = fasta.Writer(raw_fasta_file)
        new_fasta_records = []
        for rec in raw_fasta_records:
            raw_fasta_writer.write(rec)
            new_title = gene_id + "." + rec.title
            new_fasta_records.append(fasta.Record(new_title, rec.sequence))
        raw_fasta_file.close()

        return {"gene_id": gene_id,
                "chrom": self.chrom,
                "start": start,
                "end": end,
                "strand": strand,
                "num_transcripts": len(transcripts),
                "num_paths": gene_splice_graph.getNumIsoforms(),
                "num_parameters": gene_splice_graph.getNumFreeParameters(),
                "fasta_records": new_fasta_records,
                "log": log}
    
def create_psgs(pool, chrom, genes, db, max_read_length, k, unfactorized,
                ref_name, fasta_writer, index_writer):
    chrom_output_dir = os.path.join(ref_name, chrom)
    if not os.path.exists(chrom_output_dir):
        os.makedirs(chrom_output_dir)
    creator = PSGCreator(chrom, db, max_read_length, k, unfactorized,
                         chrom_output_dir)
    for result in pool.imap(creator,
                            genes.iteritems(),
                            chunksize=10):
    # for result in itertools.imap(creator,
    #                              genes.iteritems()):
        sys.stderr.write(result["log"])
        for rec in result["fasta_records"]:
            fasta_writer.write(rec)
        index_writer.write(result)
    
def main():
    usage = "usage: %prog [options] reference_name"
    parser = optparse.OptionParser(usage=usage)
    parser.add_option("-a", "--annotations", dest="gtf_filename",
                      help="GTF file containing gene/isoform annotations. "
                      "(default: GTF read from standard input)",
                      metavar="FILE")
    parser.add_option("-g", "--genome-dir", dest="genome_dir",
                      help="Directory containing chromosome FASTA files "
                      "(default: .)",
                      default=".", metavar="DIR")
    parser.add_option("-k", dest="k",
                      help="Order of the PSGs to create.  "
                      "Line graphs correspond to k=0.  "
                      "Exon graphs correspond to k=1.  "
                      "For k>1, longer range dependencies are modeled. "
                      "Set k=-1 for full-length isoform graphs. "
                      "(default: 1)",
                      default=1, type=int, metavar="INT")
    parser.add_option("-u", "--unfactorized", dest="unfactorized",
                      help="Unfactorize the PSG",
                      action="store_true")
    parser.add_option("-l", dest="max_read_length",
                      help="Maximum length of a read that will be aligned "
                      "against this reference (default: 100)",
                      default=100, type=int, metavar="INT")
    options, args = parser.parse_args()
    if len(args) != 1:
        parser.error("Invalid number of arguments")
    ref_name = args[0]

    num_cores = multiprocessing.cpu_count()
    print >>sys.stderr, "Using %d cores" % num_cores
    pool = multiprocessing.Pool(processes=num_cores)

    try:
        if not os.path.exists(ref_name):
            os.makedirs(ref_name)

        # Record maximum read length
        psg.write_reference_read_length(ref_name, options.max_read_length)
        
        # Set up file to which all PSG reference sequences will be written
        fasta_filename = os.path.join(ref_name, "references.fasta")
        fasta_file = open(fasta_filename, 'w')
        fasta_writer = fasta.Writer(fasta_file)

        # Set up file for PSG index table
        index_filename = os.path.join(ref_name, "index.txt")
        index_fields = ["gene_id",
                        "chrom",
                        "start",
                        "end",
                        "strand",
                        "num_transcripts",
                        "num_paths",
                        "num_parameters"]
        index_writer = table.FileWriter(index_filename, index_fields)
        
        db = FastaDirectoryGenomeDB(options.genome_dir, cache_last_chrom=True)

        transcripts = read_transcripts(options.gtf_filename)

        genes = transcripts_grouped_by_gene(transcripts)
        chroms = genes_grouped_by_chrom(genes)
        for chrom, genes_in_chrom in chroms.iteritems():
            psg.log("Processing chromosome %s..." % chrom)
            psg.log("\tCreating PSGs")
            create_psgs(pool, chrom, genes_in_chrom, db,
                        options.max_read_length, options.k, options.unfactorized,
                        ref_name, fasta_writer, index_writer)

        fasta_file.close()
        index_writer.close()

        # Run bowtie-build on the FASTA file
        bowtie_prefix = os.path.join(ref_name, "references")
        cmd = "bowtie-build -q %s %s" % (fasta_filename, bowtie_prefix)
        psg.log(cmd)
        psg.run_command(cmd)
        
    except psg.Error, e:
        print >>sys.stderr, "Error: %s" % e.args[0]
        sys.exit(1)
    except KeyboardInterrupt:
        print >>sys.stderr, "Caught KeyboardInterrupt, terminating workers"
        pool.terminate()
        pool.join()
        sys.exit(1)
        
if __name__ == "__main__":
    main()
